from trailblazer.apps.slurm.api import _get_slurm_queue
from trailblazer.apps.slurm.models import SqueueResult


class SlurmCLIClient:
    def __init__(self, host: str):
        self.host = host

    def get_slurm_queue(self, job_ids: str) -> SqueueResult:
        return _get_slurm_queue(job_ids=job_ids, analysis_host=self.host)



from tests.mocks.store_mock import MockStore
from trailblazer.store.database import get_tables


def test_setup_db(store: MockStore):
    """Test store contains tables."""

    # GIVEN a store which is already setup

    # THEN it should contain tables
    assert get_tables()



from http import HTTPStatus

from flask.testing import FlaskClient

from trailblazer.dto import FailedJobsRequest
from trailblazer.store.models import Analysis


def test_get_aggregated_jobs(client: FlaskClient, analysis_with_failed_job: Analysis):
    # GIVEN a valid request to aggregate failed jobs a month back
    request = FailedJobsRequest(days_back=31)
    data: str = request.model_dump_json()

    # WHEN retrieving statistics for the failed jobs
    response = client.get("/api/v1/aggregate/jobs", data=data, content_type="application/json")

    # THEN it gives a success response
    assert response.status_code == HTTPStatus.OK

    # THEN it returns the correct number of failed jobs
    failed_jobs_count: int = sum(job["count"] for job in response.json["jobs"])
    assert failed_jobs_count == 1



import logging
import os

from flask import Flask
from flask_cors import CORS
from flask_reverse_proxy import FlaskReverseProxied
from sqlalchemy.orm import scoped_session

from trailblazer.server import api, ext
from trailblazer.server.wiring import setup_dependency_injection
from trailblazer.store.database import get_session

app = Flask(__name__)
setup_dependency_injection()

LOG = logging.getLogger(__name__)

# Environment variables
SECRET_KEY = "unsafe!!!"
TEMPLATES_AUTO_RELOAD = True
SQLALCHEMY_DATABASE_URI = os.environ.get("SQLALCHEMY_DATABASE_URI", "sqlite:///:memory:")
SQLALCHEMY_POOL_RECYCLE = os.environ.get("SQLALCHEMY_POOL_RECYCLE", 7200)
SQLALCHEMY_TRACK_MODIFICATIONS = os.environ.get("FLASK_DEBUG", False)

app.config.from_object(__name__)

# register blueprints
app.register_blueprint(api.blueprint)

# configure extensions
FlaskReverseProxied(app)
cors = CORS(app, resources={r"/api/*": {"origins": "*"}}, supports_credentials=True)
ext.store.init_app(app)


@app.route("/")
def index():
    return "Welcome to Trailblazer REST API"


@app.teardown_appcontext
def teardown_session(_):
    """
    Remove the database session to ensure database resources are released when a
    request has been processed.
    """
    session: scoped_session = get_session()
    session.remove()



from trailblazer.dto import (
    AnalysesRequest,
    AnalysesResponse,
    AnalysisResponse,
    AnalysisUpdateRequest,
    CreateAnalysisRequest,
)
from trailblazer.exc import MissingAnalysis
from trailblazer.services.utils import create_analysis_response
from trailblazer.store.models import Analysis, Job
from trailblazer.store.store import Store


class AnalysisService:
    def __init__(self, store: Store):
        self.store = store

    def get_analyses(self, query: AnalysesRequest) -> AnalysesResponse:
        analyses, total_analysis_count = self.store.get_filtered_sorted_paginated_analyses(query)
        return self.create_analyses_response(
            analyses=analyses, total_analysis_count=total_analysis_count
        )

    def get_analysis(self, analysis_id: int) -> AnalysisResponse:
        if not (analysis := self.store.get_analysis_with_id(analysis_id)):
            raise MissingAnalysis(f"Analysis with id: {analysis_id} not found")
        return create_analysis_response(analysis)

    def update_analysis(self, analysis_id: int, update: AnalysisUpdateRequest) -> AnalysisResponse:
        analysis: Analysis = self.store.update_analysis(
            analysis_id=analysis_id,
            comment=update.comment,
            status=update.status,
            is_visible=update.is_visible,
        )
        return create_analysis_response(analysis)

    def add_pending_analysis(self, request_data: CreateAnalysisRequest) -> AnalysisResponse:
        analysis: Analysis = self.store.add_pending_analysis(request_data)
        return create_analysis_response(analysis)

    def create_analyses_response(
        self, analyses: list[Analysis], total_analysis_count: int
    ) -> AnalysesResponse:
        response_data: list[dict] = []
        for analysis in analyses:
            analysis_data = analysis.to_dict()
            failed_job: Job = self.store.get_latest_failed_job_for_analysis(analysis.id)
            analysis_data["failed_job"] = failed_job.to_dict() if failed_job else None
            response_data.append(analysis_data)
        return AnalysesResponse(analyses=response_data, total_count=total_analysis_count)

    def update_ongoing_analyses(self) -> None:
        self.store.update_ongoing_analyses()



from pydantic import BaseModel


class SlurmError(BaseModel):
    description: str | None = None
    error_number: int | None = None
    error: str | None = None
    source: str | None = None


class SlurmWarning(BaseModel):
    description: str | None = None
    source: str | None = None


class NumberWithFlags(BaseModel):
    set: bool | None = None
    infinite: bool | None = None
    number: int | None = None


class JobResources(BaseModel):
    nodes: str | None = None
    allocated_cores: int | None = None
    allocated_cpus: int | None = None
    allocated_hosts: int | None = None
    allocated_nodes: list[str] | None = None


class ExitCodeSignal(BaseModel):
    id: NumberWithFlags | None = None
    name: str | None = None


class ExitCode(BaseModel):
    status: list[str] | None = None
    return_code: NumberWithFlags | None = None
    signal: ExitCodeSignal | None = None


class SlurmJobInfo(BaseModel):
    account: str | None = None
    accrue_time: NumberWithFlags | None = None
    admin_comment: str | None = None
    allocating_node: str | None = None
    array_job_id: NumberWithFlags | None = None
    array_task_id: NumberWithFlags | None = None
    array_max_tasks: NumberWithFlags | None = None
    array_task_string: str | None = None
    association_id: int | None = None
    batch_features: str | None = None
    batch_flag: bool | None = None
    batch_host: str | None = None
    flags: list[str] | None = None
    burst_buffer: str | None = None
    burst_buffer_state: str | None = None
    cluster: str | None = None
    cluster_features: str | None = None
    command: str | None = None
    comment: str | None = None
    container: str | None = None
    container_id: str | None = None
    contiguous: bool | None = None
    core_spec: int | None = None
    thread_spec: int | None = None
    cores_per_socket: NumberWithFlags | None = None
    billable_tres: NumberWithFlags | None = None
    cpus_per_task: NumberWithFlags | None = None
    cpu_frequency_minimum: NumberWithFlags | None = None
    cpu_frequency_maximum: NumberWithFlags | None = None
    cpu_frequency_governor: NumberWithFlags | None = None
    cpus_per_tres: str | None = None
    cron: str | None = None
    deadline: NumberWithFlags | None = None
    delay_boot: NumberWithFlags | None = None
    dependency: str | None = None
    derived_exit_code: ExitCode | None = None
    eligible_time: NumberWithFlags | None = None
    end_time: NumberWithFlags | None = None
    excluded_nodes: str | None = None
    exit_code: ExitCode | None = None
    extra: str | None = None
    failed_node: str | None = None
    features: str | None = None
    federation_origin: str | None = None
    federation_siblings_active: str | None = None
    federation_siblings_viable: str | None = None
    gres_detail: list[str] | None = None
    group_id: int | None = None
    group_name: str | None = None
    het_job_id: NumberWithFlags | None = None
    het_job_id_set: str | None = None
    het_job_offset: NumberWithFlags | None = None
    job_id: int | None = None
    job_resources: JobResources | None = None
    job_size_str: list[str] | None = None
    job_state: list[str] | None = None
    last_sched_evaluation: NumberWithFlags | None = None
    licenses: str | None = None
    mail_type: list[str] | None = None
    mail_user: str | None = None
    max_cpus: NumberWithFlags | None = None
    max_nodes: NumberWithFlags | None = None
    mcs_label: str | None = None
    memory_per_tres: str | None = None
    name: str | None = None
    network: str | None = None
    nodes: str | None = None
    nice: int | None = None
    tasks_per_core: NumberWithFlags | None = None
    tasks_per_tres: NumberWithFlags | None = None
    tasks_per_node: NumberWithFlags | None = None
    tasks_per_socket: NumberWithFlags | None = None
    tasks_per_board: NumberWithFlags | None = None
    cpus: NumberWithFlags | None = None
    node_count: NumberWithFlags | None = None
    tasks: NumberWithFlags | None = None
    partition: str | None = None
    prefer: str | None = None
    memory_per_cpu: NumberWithFlags | None = None
    memory_per_node: NumberWithFlags | None = None
    minimum_cpus_per_node: NumberWithFlags | None = None
    minimum_tmp_disk_per_node: NumberWithFlags | None = None
    power: list[str] | None = None
    preempt_time: NumberWithFlags | None = None
    preemptable_time: NumberWithFlags | None = None
    pre_sus_time: NumberWithFlags | None = None
    priority: NumberWithFlags | None = None
    profile: list[str] | None = None
    qos: str | None = None
    reboot: bool | None = None
    required_nodes: str | None = None
    minimum_switches: int | None = None
    requeue: bool | None = None
    resize_time: NumberWithFlags | None = None
    restart_cnt: int | None = None
    resv_name: str | None = None
    scheduled_nodes: str | None = None
    selinux_context: str | None = None
    shared: list[str] | None = None
    exclusive: list[str] | None = None
    oversubscribe: bool | None = None
    show_flags: list[str] | None = None
    sockets_per_board: int | None = None
    sockets_per_node: NumberWithFlags | None = None
    start_time: NumberWithFlags | None = None
    state_description: str | None = None
    state_reason: str | None = None
    standard_error: str | None = None
    standard_input: str | None = None
    standard_output: str | None = None
    submit_time: NumberWithFlags | None = None
    suspend_time: NumberWithFlags | None = None
    system_comment: str | None = None
    time_limit: NumberWithFlags | None = None
    time_minimum: NumberWithFlags | None = None
    threads_per_core: NumberWithFlags | None = None
    tres_bind: str | None = None
    tres_freq: str | None = None
    tres_per_job: str | None = None
    tres_per_node: str | None = None
    tres_per_socket: str | None = None
    tres_per_task: str | None = None
    tres_req_str: str | None = None
    tres_alloc_str: str | None = None
    user_id: int | None = None
    user_name: str | None = None
    maximum_switch_wait_time: int | None = None
    wckey: str | None = None
    current_working_directory: str | None = None



from abc import ABC, abstractmethod

from trailblazer.services.slurm.dtos import SlurmJobInfo


class SlurmService(ABC):
    @abstractmethod
    def get_job_info(self, job_id: int) -> SlurmJobInfo:
        pass



import os
from dependency_injector import containers, providers

from trailblazer.clients.slurm_cli_client.slurm_cli_client import SlurmCLIClient
from trailblazer.services.analysis_service import AnalysisService
from trailblazer.services.job_service import JobService
from trailblazer.services.slurm.slurm_cli_service.slurm_cli_service import SlurmCLIService
from trailblazer.store.store import Store


class Container(containers.DeclarativeContainer):
    slurm_host: str | None = os.environ.get("ANALYSIS_HOST")

    store = providers.Singleton(Store)
    slurm_client = providers.Singleton(SlurmCLIClient, host=slurm_host)
    slurm_service = providers.Singleton(SlurmCLIService, client=slurm_client)

    job_service = providers.Factory(JobService, store=store, slurm_service=slurm_service)
    analysis_service = providers.Factory(AnalysisService, store=store)



from datetime import datetime
import pytest

from trailblazer.clients.slurm_cli_client.slurm_cli_client import SlurmCLIClient
from trailblazer.constants import TrailblazerStatus
from trailblazer.services.job_service import JobService
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_cli_service.slurm_cli_service import SlurmCLIService

from trailblazer.store.store import Store


@pytest.fixture
def upload_job_info() -> SlurmJobInfo:
    return SlurmJobInfo(
        slurm_id=123,
        name="upload",
        status=TrailblazerStatus.COMPLETED,
        started_at=datetime.now(),
        elapsed=100,
    )


@pytest.fixture
def job_service(analysis_store: Store):
    slurm_client = SlurmCLIClient("host")
    slurm_service = SlurmCLIService(slurm_client)
    return JobService(slurm_service=slurm_service, store=analysis_store)



from pathlib import Path

from trailblazer.constants import FileFormat
from trailblazer.io.controller import ReadFile, ReadStream


def test_get_content_from_file(example_yaml_path: Path):
    """
    Tests get content from file.
    """
    # GIVEN a YAML file

    # WHEN reading the YAML file
    yaml_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=example_yaml_path
    )

    # Then assert a dict is returned
    assert isinstance(yaml_content, dict)


def test_get_content_from_file_when_json(example_json_path: Path):
    """
    Tests get content from file when JSON.
    """
    # GIVEN a JSON file

    # WHEN reading the JSON file
    json_content: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=example_json_path
    )

    # Then assert a dict is returned
    assert isinstance(json_content, dict)


def test_get_content_from_stream_when_csv(csv_stream: str):
    """Tests read a CSV stream."""
    # GIVEN a string in CSV format

    # WHEN reading the CSV content in string
    raw_csv_content: list = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV, stream=csv_stream
    )

    # THEN assert that a list is returned
    assert isinstance(raw_csv_content, list)


def test_get_content_from_stream_when_csv_when_dict(csv_stream: str):
    """Tests read a CSV stream into a dict."""
    # GIVEN a string in CSV format

    # WHEN reading the CSV content in string
    raw_csv_content: list = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=csv_stream,
        read_to_dict=True,
    )

    # Then assert a list is returned and that the first element is a dict
    assert isinstance(raw_csv_content, list)
    assert isinstance(raw_csv_content[0], dict)



from trailblazer.apps.slurm.models import SqueueJob
from trailblazer.services.slurm.dtos import SlurmJobInfo


def create_job_info_dto(job: SqueueJob) -> SlurmJobInfo:
    return SlurmJobInfo(
        name=job.step,
        slurm_id=job.id,
        status=job.status,
        started_at=job.started_at,
        elapsed=job.time_elapsed,
    )



from datetime import datetime

from sqlalchemy.orm import Query

from tests.mocks.store_mock import MockStore
from trailblazer.constants import TrailblazerStatus
from trailblazer.store.models import Analysis, Info, Job


def test_info_query(info_store: MockStore):
    """Test base model table query."""

    # GIVEN a store

    # WHEN getting an object using the query
    info: Info = info_store.get_query(table=Info).first()

    # THEN it should return an Info object with a default "created_at" date
    assert isinstance(info.created_at, datetime)


def test_get_job_query_with_name_and_count_labels(job_store: MockStore):
    """Test Job query with labels and a count."""

    # GIVEN a job store

    # WHEN getting an object using the query
    category: Query = job_store.get_job_query_with_name_and_count_labels()

    # THEN it should return an Job object with a name and count
    assert isinstance(category, Query)

    # WHEN reading the database
    category = category.first()

    # THEN the returned object should have a name and count attribute
    assert category.name == "0"

    assert category.count == job_store.get_query(table=Job).count()


def test_get_analyses_query_by_search_term_and_is_visible(analysis_store: MockStore):
    """Test return analyses query by search term and is visible"""

    # GIVEN a store
    analyses_query: Query = analysis_store.get_query(table=Analysis)

    # WHEN getting an object using the query
    analyses: Query = analysis_store.get_analyses_query_by_search(
        search_term=TrailblazerStatus.PENDING, analyses=analyses_query
    )

    # THEN it should return a query
    assert isinstance(analyses, Query)

    # THEN it should return an object with the search term
    assert analyses.first().status == TrailblazerStatus.PENDING



from datetime import datetime
from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query
from trailblazer.constants import JobType

from trailblazer.store.models import Job


def filter_jobs_by_since_when(jobs: Query, since_when: datetime, **kwargs) -> Query:
    """Filter jobs before the supplied date."""
    return jobs.filter(Job.started_at > since_when)


def filter_jobs_by_status(jobs: Query, status: str, **kwargs) -> Query:
    """Filter jobs with matching status."""
    return jobs.filter(Job.status == status)


def filter_jobs_by_statuses(jobs: Query, statuses: list[str], **kwargs) -> Query:
    return jobs.filter(Job.status.in_(statuses))


def filter_jobs_by_analysis_id(jobs: Query, analysis_id: str, **kwargs) -> Query:
    """Filter jobs with matching analysis id."""
    return jobs.filter(Job.analysis_id == analysis_id)


def sort_jobs_by_started_at(jobs: Query, **kwargs) -> Query:
    """Sort jobs by start date."""
    return jobs.order_by(Job.started_at.desc())


def filter_jobs_by_id(jobs: Query, job_id: int, **kwargs) -> Query:
    """Filter jobs with matching id."""
    return jobs.filter(Job.id == job_id)


def filter_by_job_type(jobs: Query, job_type: str, **kwargs):
    return jobs.filter(Job.job_type == job_type)


class JobFilter(Enum):
    """Define Job filter functions."""

    FILTER_BY_SINCE_WHEN: Callable = filter_jobs_by_since_when
    FILTER_BY_STATUS: Callable = filter_jobs_by_status
    FILTER_BY_STATUSES: Callable = filter_jobs_by_statuses
    FILTER_BY_ANALYSIS_ID: Callable = filter_jobs_by_analysis_id
    FILTER_BY_ID: Callable = filter_jobs_by_id
    FILTER_BY_TYPE: Callable = filter_by_job_type
    SORT_BY_STARTED_AT: Callable = sort_jobs_by_started_at


def apply_job_filters(
    jobs: Query,
    filters: list[Callable],
    since_when: datetime | None = None,
    status: str | None = None,
    statuses: list[str] | None = None,
    analysis_id: str | None = None,
    job_id: int | None = None,
    job_type: str | None = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filters:
        jobs: Query = function(
            jobs=jobs,
            since_when=since_when,
            status=status,
            statuses=statuses,
            analysis_id=analysis_id,
            job_id=job_id,
            job_type=job_type,
        )
    return jobs



from datetime import datetime
from pydantic import BaseModel

from trailblazer.constants import SlurmJobStatus


class SlurmJobInfo(BaseModel):
    slurm_id: int
    name: str
    status: SlurmJobStatus
    started_at: datetime
    elapsed: int



import logging

from trailblazer.store.models import User

LOG = logging.getLogger(__name__)


def is_existing_user(user: User, email: str) -> bool:
    """Check if supplied user exist in db."""
    if not user:
        LOG.error(f"User with email {email} not found")
        return False
    LOG.info(f"User with email {email} found")
    return True


def is_user_archived(user: User, email: str) -> bool:
    """Check if user is archived."""
    LOG.info(f"User with email {email} has archive status: {user.is_archived}")
    return user.is_archived



import requests

from trailblazer.clients.slurm_api_client.dto import (
    SlurmCancelJobResponse,
    SlurmJobResponse,
    SlurmJobsResponse,
)


class SlurmApiClient:
    def __init__(self, base_url: str, access_token: str, user_name: str) -> None:
        self.base_url = base_url
        self.headers = {"X-SLURM-USER-NAME": user_name, "X-SLURM-USER-TOKEN": access_token}

    def get_job(self, job_id: str) -> SlurmJobResponse:
        endpoint: str = f"{self.base_url}/slurm/v0.0.40/job/{job_id}"
        response = requests.get(endpoint, headers=self.headers)
        response.raise_for_status()
        return SlurmJobResponse.model_validate(response.json())

    def get_jobs(self) -> SlurmJobsResponse:
        endpoint: str = f"{self.base_url}/slurm/v0.0.40/jobs"
        response = requests.get(endpoint, headers=self.headers)
        response.raise_for_status()
        return SlurmJobsResponse.model_validate(response.json())

    def cancel_job(self, job_id: str) -> SlurmCancelJobResponse:
        endpoint: str = f"{self.base_url}/slurm/v0.0.40/job/{job_id}"
        response = requests.delete(endpoint, headers=self.headers)
        response.raise_for_status()
        return SlurmCancelJobResponse.model_validate(response.json())



from datetime import datetime
import logging

from trailblazer.dto import CreateJobRequest, FailedJobsRequest, FailedJobsResponse, JobResponse
from trailblazer.services.slurm.dtos import SlurmJobInfo
from trailblazer.services.slurm.slurm_service import SlurmService
from trailblazer.services.utils import create_job_response, create_failed_jobs_response
from trailblazer.store.models import Job
from trailblazer.store.store import Store
from trailblazer.utils.datetime import get_date_number_of_days_ago

LOG = logging.getLogger(__name__)


class JobService:
    def __init__(self, store: Store, slurm_service: SlurmService):
        self.store = store
        self.slurm_service = slurm_service

    def get_failed_jobs(self, request: FailedJobsRequest) -> FailedJobsResponse:
        time_window: datetime = get_date_number_of_days_ago(request.days_back)
        failed_jobs: list[dict] = self.store.get_failed_jobs_stats(time_window)
        return create_failed_jobs_response(failed_jobs)

    def add_job(self, analysis_id: int, data: CreateJobRequest) -> JobResponse:
        job: Job = self.store.add_job(analysis_id=analysis_id, job_request=data)
        return create_job_response(job)

    def update_upload_jobs(self) -> None:
        jobs: list[Job] = self.store.get_ongoing_upload_jobs()
        for job in jobs:
            LOG.info(f"Updating upload job {job.id}")
            updated_job: SlurmJobInfo = self.slurm_service.get_job_info(job.slurm_id)
            self.store.update_job(job_id=job.id, job_info=updated_job)



import subprocess
from datetime import datetime

import pytest
from click.testing import CliRunner

import trailblazer
from tests.mocks.store_mock import MockStore
from trailblazer.cli.core import (
    add_user_to_db,
    archive_user,
    base,
    cancel,
    delete,
    get_user_from_db,
    get_users_from_db,
    ls_cmd,
    scan,
    set_analysis_completed,
    set_analysis_status,
    unarchive_user,
)
from trailblazer.constants import CharacterFormat, SlurmJobStatus, TrailblazerStatus
from trailblazer.containers import Container
from trailblazer.store.models import Analysis

FUNC_GET_SLURM_SQUEUE_OUTPUT_PATH: str = "trailblazer.store.crud.update.get_slurm_squeue_output"


def test_base(cli_runner):
    # GIVEN the cli

    # WHEN asking for the version
    result = cli_runner.invoke(base, ["--version"])

    # THEN it should print the name and version of the tool only
    assert trailblazer.__title__ in result.output
    assert trailblazer.__version__ in result.output


def test_set_analysis_completed(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    failed_analysis_case_id: str,
    process_exit_success: int,
):
    """Test setting an analysis to status complete."""
    # GIVEN an analysis with status failed
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
    analysis: Analysis | None = trailblazer_db.get_latest_analysis_for_case(
        case_id=failed_analysis_case_id
    )

    # Make sure status is not "completed"
    assert analysis.status != TrailblazerStatus.COMPLETED

    # WHEN running command
    result = cli_runner.invoke(set_analysis_completed, [str(analysis.id)], obj=trailblazer_context)

    # THEN command runs successfully
    assert result.exit_code == process_exit_success

    # THEN status will be set to "complete"
    analysis: Analysis | None = trailblazer_db.get_latest_analysis_for_case(
        case_id=failed_analysis_case_id
    )
    assert analysis.status == TrailblazerStatus.COMPLETED


def test_set_analysis_status(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    failed_analysis_case_id: str,
    process_exit_success: int,
):
    """Test that the latest analysis status is updated for a given case id."""

    # GIVEN an analysis with status failed
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
    analysis: Analysis | None = trailblazer_db.get_latest_analysis_for_case(
        case_id=failed_analysis_case_id
    )

    # Make sure status is not "qc"
    assert analysis.status != TrailblazerStatus.QC

    # WHEN running command
    result = cli_runner.invoke(
        set_analysis_status, ["--status", "qc", failed_analysis_case_id], obj=trailblazer_context
    )

    # THEN command runs successfully
    assert result.exit_code == process_exit_success

    # THEN status will be set to "qc
    analysis = trailblazer_db.get_latest_analysis_for_case(case_id=failed_analysis_case_id)
    assert analysis.status == TrailblazerStatus.QC


def test_set_analysis_status_error(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    failed_analysis_case_id: str,
):
    """Test that setting the status to a non-accepted value raises an error."""

    # GIVEN an analysis with status failed

    # WHEN running command
    result = cli_runner.invoke(
        set_analysis_status,
        ["--status", "non_existing_status", failed_analysis_case_id],
        obj=trailblazer_context,
    )

    # THEN an error should be raised
    assert result.exit_code != 0
    assert "Invalid status" in caplog.text


def test_cancel_non_existent_analysis_id(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    process_exit_success: int,
):
    """Test cancelling an analysis using a non existing analysis id."""
    caplog.set_level("ERROR")

    # GIVEN Trailblazer database with analyses and jobs
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
    trailblazer_db.update_ongoing_analyses()

    # WHEN running cancel on non-existing entry
    result = cli_runner.invoke(cancel, ["123"], obj=trailblazer_context)

    # THEN command runs successfully
    assert result.exit_code == process_exit_success

    # THEN error log informs user that analysis does not exist
    assert "does not exist" in caplog.text


def test_cancel_not_running(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    failed_analysis_case_id: str,
    mocker,
    process_exit_success: int,
    slurm_squeue_output: dict[str, str],
):
    """Test cancelling an analysis, which is not running."""
    caplog.set_level("ERROR")

    # GIVEN SLURM squeue output for an analysis
    mocker.patch(
        FUNC_GET_SLURM_SQUEUE_OUTPUT_PATH,
        return_value=subprocess.check_output(
            ["cat", slurm_squeue_output.get(failed_analysis_case_id)]
        ).decode(CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8),
    )

    # GIVEN an analysis that is NOT running
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
    analysis: Analysis | None = trailblazer_db.get_latest_analysis_for_case(
        case_id=failed_analysis_case_id
    )
    trailblazer_db.update_ongoing_analyses()

    # WHEN trying to cancel analysis
    result = cli_runner.invoke(cancel, [str(analysis.id)], obj=trailblazer_context)

    # THEN command runs successfully
    assert result.exit_code == process_exit_success

    # THEN error log informs user that analysis cannot be cancelled as it is not running
    assert "is not running" in caplog.text


def test_cancel_with_ongoing_analysis(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    mocker,
    ongoing_analysis_case_id: str,
    slurm_squeue_output: dict[str, str],
    process_exit_success: int,
):
    """Test all ongoing analysis jobs are cancelled."""
    # GIVEN SLURM squeue output for an analysis
    mocker.patch(
        FUNC_GET_SLURM_SQUEUE_OUTPUT_PATH,
        return_value=subprocess.check_output(
            ["cat", slurm_squeue_output.get(ongoing_analysis_case_id)]
        ).decode(CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8),
    )

    # GIVEN SLURM scancel output for an analysis
    mocker.patch("trailblazer.store.crud.update.cancel_slurm_job", return_value=None)

    caplog.set_level("INFO")

    # GIVEN an ongoing analysis
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
    trailblazer_db.update_ongoing_analyses()
    analysis: Analysis | None = trailblazer_db.get_latest_analysis_for_case(
        case_id=ongoing_analysis_case_id
    )

    # Analysis should have jobs that can be cancelled
    assert analysis.jobs

    # WHEN running cancel command
    result = cli_runner.invoke(cancel, [str(analysis.id)], obj=trailblazer_context)

    # THEN command should run successfully
    assert result.exit_code == process_exit_success

    # THEN log should inform of successful cancellation
    assert "cancelled successfully" in caplog.text
    assert "Cancelling" in caplog.text

    # THEN job id from squeue output will be cancelled
    assert f"{analysis.jobs[3].slurm_id}" in caplog.text

    # THEN analysis status is set to cancelled
    assert SlurmJobStatus.CANCELLED in analysis.comment
    assert analysis.status == TrailblazerStatus.CANCELLED


def test_delete_nonexisting(
    cli_runner: CliRunner, trailblazer_context: dict[str, MockStore], caplog
):
    with caplog.at_level("ERROR"):
        # GIVEN Trailblazer database with analyses and jobs
        trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
        trailblazer_db.update_ongoing_analyses()

        # WHEN trying to delete analysis not in database
        cli_runner.invoke(delete, ["123"], obj=trailblazer_context)

        # THEN error log informs user that analysis is not found
        assert "Analysis not found" in caplog.text


def test_delete_ongoing_fail(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    ongoing_analysis_case_id: str,
    process_exit_success: int,
):
    """Test deleting an ongoing analysis without using the force flag."""
    caplog.set_level("ERROR")

    # GIVEN Trailblazer database with analyses and jobs
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
    trailblazer_db.update_ongoing_analyses()

    # GIVEN an analysis that is ongoing
    analysis: Analysis | None = trailblazer_db.get_latest_analysis_for_case(
        case_id=ongoing_analysis_case_id
    )

    # WHEN trying to delete ongoing analysis without --force flag
    result = cli_runner.invoke(delete, [str(analysis.id)], obj=trailblazer_context)

    # THEN command executes successfully
    assert result.exit_code == process_exit_success

    # THEN error log informs user that ongoing analysis cannot be deleted without --force flag
    assert "--force" in caplog.text

    # THEN analysis should not be cancelled
    assert analysis.status != TrailblazerStatus.CANCELLED


def test_delete_ongoing_force(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    ongoing_analysis_case_id,
    process_exit_success: int,
):
    caplog.set_level("INFO")

    # GIVEN Trailblazer database with analyses and jobs and an ongoing analysis
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]
    trailblazer_db.update_ongoing_analyses()
    analysis: Analysis | None = trailblazer_db.get_latest_analysis_for_case(
        case_id=ongoing_analysis_case_id
    )

    # WHEN running command with --force flag
    result = cli_runner.invoke(delete, [str(analysis.id), "--force"], obj=trailblazer_context)

    # THEN command runs successfully
    assert result.exit_code == process_exit_success

    # THEN log informs user that Trailblazer is deleting analysis
    assert "Deleting" in caplog.text

    # THEN the analysis should have been deleted from the database
    assert not trailblazer_db.get_latest_analysis_for_case(case_id=analysis.case_id)


def test_get_user_not_in_database(
    cli_runner: CliRunner, trailblazer_context: dict[str, MockStore], caplog
):
    """Test getting user when user is not in the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("ERROR")

    # WHEN looking for a user that is not added
    cli_runner.invoke(get_user_from_db, ["not_in_db@disco.com"], obj=trailblazer_context)

    # THEN log should inform that user was not found
    assert "not found" in caplog.text


def test_get_user_in_database(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    user_email: str,
    username: str,
):
    """Test getting user when user is in the database."""
    # GIVEN populated Trailblazer database
    caplog.set_level("INFO")

    # WHEN looking for a user that is added
    cli_runner.invoke(get_user_from_db, [user_email], obj=trailblazer_context)

    # THEN log informs the user about their full name
    assert username in caplog.text


def test_add_user(cli_runner: CliRunner, trailblazer_context: dict[str, MockStore], caplog):
    """Test adding a user to the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN adding new user
    cli_runner.invoke(
        add_user_to_db, ["harriers@disco.com", "--name", "Harry DuBois"], obj=trailblazer_context
    )

    # THEN log should inform that user is added
    assert "New user added" in caplog.text


def test_add_user_when_already_exists(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    user_email: str,
    username: str,
):
    """Test adding a user to the database when the user already exists."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN adding new user
    cli_runner.invoke(add_user_to_db, [user_email, "--name", username], obj=trailblazer_context)

    # THEN log should inform that user is added
    assert f"User with email {user_email} found" in caplog.text


def test_users(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    user_email: str,
    username: str,
):
    """Test getting users from the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN getting users
    cli_runner.invoke(get_users_from_db, ["--email", user_email], obj=trailblazer_context)

    # THEN log shows users
    assert username in caplog.text


def test_archive_user(
    cli_runner: CliRunner, trailblazer_context: dict[str, MockStore], caplog, user_email: str
):
    """Test archiving a user in the database."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN archiving user
    cli_runner.invoke(archive_user, [user_email], obj=trailblazer_context)

    # THEN log shows users as archived
    assert f"User archived: {user_email}" in caplog.text


def test_archive_user_when_already_archived(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    archived_user_email: str,
):
    """Test archiving a user in the database when already archived."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN archiving user
    cli_runner.invoke(archive_user, [archived_user_email], obj=trailblazer_context)

    # THEN log shows users is already archived
    assert "has archive status: True" in caplog.text


def test_archive_user_when_non_existing_user(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    archived_user_email: str,
):
    """Test archiving a user in the database when user does not exist."""
    # GIVEN populated Trailblazer database

    # GIVEN a user email that does not exist
    non_existing_email: str = "does-not.exist@none.com"

    # WHEN archiving user
    cli_runner.invoke(archive_user, [non_existing_email], obj=trailblazer_context)

    # THEN log shows users is not found
    assert f"User with email {non_existing_email} not found" in caplog.text


def test_unarchive_user_when_not_archived(
    cli_runner: CliRunner, trailblazer_context: dict[str, MockStore], caplog, user_email: str
):
    """Test unarchiving a user in the database which is not archived."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN unarchiving user
    cli_runner.invoke(unarchive_user, [user_email], obj=trailblazer_context)

    # THEN log shows users is not archived
    assert f"User with email {user_email} has archive status: False" in caplog.text


def test_unarchive_user(
    cli_runner: CliRunner,
    trailblazer_context: dict[str, MockStore],
    caplog,
    archived_user_email: str,
):
    """Test unarchiving a user in the database which is archived."""
    # GIVEN populated Trailblazer database

    caplog.set_level("INFO")

    # WHEN unarchiving user
    cli_runner.invoke(unarchive_user, [archived_user_email], obj=trailblazer_context)

    # THEN log shows users as not archived
    assert f"User unarchived: {archived_user_email}" in caplog.text


def test_scan(
    cli_runner: CliRunner,
    analysis_store: MockStore,
    caplog,
    mocker,
    ongoing_analysis_case_id: str,
    slurm_squeue_output: dict[str, str],
):
    """Test scanning for analyses and updating analysis status."""
    caplog.set_level("INFO")

    # GIVEN SLURM squeue output for an analysis
    mocker.patch(
        FUNC_GET_SLURM_SQUEUE_OUTPUT_PATH,
        return_value=subprocess.check_output(
            ["cat", slurm_squeue_output.get(ongoing_analysis_case_id)]
        ).decode(CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8),
    )

    # GIVEN populated Trailblazer database with pending analyses

    # WHEN running trailblazer scan command
    cli_runner.invoke(scan, [])

    # THEN log that analyses are updated
    assert "All analyses updated" in caplog.text

    # THEN the status of analysis should be updated from pending
    analysis = analysis_store.get_latest_analysis_for_case(ongoing_analysis_case_id)
    assert analysis.status == TrailblazerStatus.RUNNING


@pytest.mark.parametrize(
    "case_id, status",
    [
        ("blazinginsect", TrailblazerStatus.RUNNING),
        ("crackpanda", TrailblazerStatus.FAILED),
        ("fancymole", TrailblazerStatus.COMPLETED),
        ("happycow", TrailblazerStatus.PENDING),
    ],
)
def test_ls(
    case_id: str,
    cli_runner: CliRunner,
    mocker,
    process_exit_success: int,
    slurm_squeue_output: dict[str, str],
    status: str,
    timestamp_now: datetime,
    trailblazer_context: dict[str, MockStore],
):
    """Test the Trailblazer ls CLI command using different cases and statuses."""
    # GIVEN SLURM squeue output for an analysis
    mocker.patch(
        FUNC_GET_SLURM_SQUEUE_OUTPUT_PATH,
        return_value=subprocess.check_output(["cat", slurm_squeue_output.get(case_id)]).decode(
            CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8
        ),
    )
    # GIVEN a populated Trailblazer database with pending analyses
    trailblazer_db: MockStore = trailblazer_context["trailblazer_db"]

    # Update analyses to their expected status
    trailblazer_db.update_ongoing_analyses()

    # WHEN running ls command while filtering by status
    result = cli_runner.invoke(
        ls_cmd, ["--status", status, "--before", str(timestamp_now.date())], obj=trailblazer_context
    )

    # THEN command runs successfully with provided options
    assert result.exit_code == process_exit_success

    # THEN ls print info about cases with that status
    assert case_id in result.output



from flask import Flask
from trailblazer.store.database import initialize_database
from trailblazer.store.store import Store


class FlaskStore(Store):
    def __init__(self, app=None):
        if app:
            self.init_app(app)

    def init_app(self, app: Flask):
        uri = app.config["SQLALCHEMY_DATABASE_URI"]
        initialize_database(uri)
        super(FlaskStore, self).__init__()


store = FlaskStore()



import logging
import sys
from datetime import datetime
from pathlib import Path

import click
import coloredlogs
from sqlalchemy.orm import scoped_session
from dependency_injector.wiring import inject, Provide

import trailblazer
from trailblazer.cli.utils.ls_helper import _get_ls_analysis_message
from trailblazer.cli.utils.user_helper import is_existing_user, is_user_archived
from trailblazer.constants import TRAILBLAZER_TIME_STAMP, FileFormat, TrailblazerStatus
from trailblazer.containers import Container
from trailblazer.environ import environ_email
from trailblazer.io.controller import ReadFile
from trailblazer.models import Config
from trailblazer.server.wiring import setup_dependency_injection
from trailblazer.services.analysis_service import AnalysisService
from trailblazer.services.job_service import JobService
from trailblazer.store.database import get_session, initialize_database
from trailblazer.store.models import Analysis, User
from trailblazer.store.store import Store

LOG = logging.getLogger(__name__)
LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR"]


class DatabaseResource:
    """
    Setup the database and ensure resources are released when the
    CLI command has been processed.
    """

    def __init__(self, db_uri: str):
        self.db_uri = db_uri

    def __enter__(self):
        initialize_database(self.db_uri)

    def __exit__(self, _, __, ___):
        session: scoped_session = get_session()
        session.remove()


@click.group()
@click.option("-c", "--config", required=True, type=click.File())
@click.option(
    "-l", "--log-level", type=click.Choice(LEVELS), default="INFO", help="lowest level to log at"
)
@click.option("--verbose", is_flag=True, help="Show full log information, time stamp etc")
@click.version_option(trailblazer.__version__, prog_name=trailblazer.__title__)
@click.pass_context
def base(
    context: click.Context,
    config: click.File,
    log_level: str,
    verbose: bool,
):
    """Trailblazer - Monitor analyses"""
    if verbose:
        log_format = "%(asctime)s %(hostname)s %(name)s[%(process)d] %(levelname)s %(message)s"
    else:
        log_format = "%(message)s" if sys.stdout.isatty() else None

    coloredlogs.install(level=log_level, fmt=log_format)
    setup_dependency_injection()

    validated_config = Config(
        **ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=Path(config.name))
    )
    context.obj = dict(validated_config)
    context.with_resource(DatabaseResource(validated_config.database_url))
    context.obj["trailblazer_db"] = Store()


@base.command()
@click.option("--reset", is_flag=True, help="reset database before setting up tables")
@click.option("--force", is_flag=True, help="bypass manual confirmations")
@click.pass_context
def init(context, reset, force):
    """Setup the database."""
    existing_tables = context.obj["trailblazer_db"].engine.table_names()
    if force or reset:
        if existing_tables and not force:
            message = f"Delete existing tables? [{', '.join(existing_tables)}]"
            click.confirm(click.style(message, fg="yellow"), abort=True)
        context.obj["trailblazer_db"].drop_all()
    elif existing_tables:
        LOG.warning("Database already exists, use '--reset'")
        context.abort()
    context.obj["trailblazer_db"].setup()
    LOG.info(f"Success! New tables: {', '.join(context.obj['trailblazer'].engine.table_names())}")


@base.command()
@inject
def scan(
    job_service: JobService = Provide[Container.job_service],
    analysis_service: AnalysisService = Provide[Container.analysis_service],
):
    """Scan ongoing analyses in SLURM"""
    analysis_service.update_ongoing_analyses()
    job_service.update_upload_jobs()
    LOG.info("All analyses updated!")


@base.command("update-analysis")
@click.argument("analysis_id")
@click.pass_context
def update_analysis(context, analysis_id: int):
    """Update status of a single analysis."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    trailblazer_db.update_run_status(analysis_id=analysis_id)


@base.command("add-user")
@click.option("--name", help="Name of new user to add")
@click.argument("email", default=environ_email())
@click.pass_context
def add_user_to_db(context, email: str, name: str) -> None:
    """Add a new user to the database."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    existing_user = trailblazer_db.get_user(email=email, exclude_archived=False)
    if is_existing_user(user=existing_user, email=email):
        return
    new_user = trailblazer_db.add_user(email=email, name=name)
    LOG.info(f"New user added: {new_user.email} ({new_user.id})")


@base.command("get-user")
@click.argument("email", default=environ_email())
@click.pass_context
def get_user_from_db(context, email: str) -> None:
    """Display information about an existing user."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    existing_user = trailblazer_db.get_user(email=email, exclude_archived=False)
    if not is_existing_user(user=existing_user, email=email):
        return
    LOG.info(f"Existing user found: {existing_user.to_dict()}")


@base.command("get-users")
@click.option("--name", type=click.types.STRING, help="Name of new users to list")
@click.option("--email", type=click.types.STRING, help="Name of new users to list")
@click.option("--exclude-archived", is_flag=True, help="Exclude archived users")
@click.pass_context
def get_users_from_db(context, name: str, email: str, exclude_archived: bool) -> None:
    """Display information about existing users."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    users: list[User] = trailblazer_db.get_users(
        email=email, exclude_archived=exclude_archived, name=name
    )
    LOG.info("Listing users in database:")
    for user in users:
        LOG.info(f"{user.to_dict()}")


@base.command("archive-user")
@click.argument("email", default=environ_email())
@click.pass_context
def archive_user(context, email: str) -> None:
    """Archive an existing user identified by email."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    existing_user: User = trailblazer_db.get_user(email=email, exclude_archived=False)

    if not is_existing_user(user=existing_user, email=email):
        return
    if is_user_archived(user=existing_user, email=email):
        return

    trailblazer_db.update_user_is_archived(user=existing_user, archive=True)
    LOG.info(f"User archived: {email}")


@base.command("unarchive-user")
@click.argument("email", default=environ_email())
@click.pass_context
def unarchive_user(context, email: str) -> None:
    """Unarchive an existing user identified by email."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    existing_user: User = trailblazer_db.get_user(email=email, exclude_archived=False)

    if not is_existing_user(user=existing_user, email=email):
        return
    if not is_user_archived(user=existing_user, email=email):
        return

    trailblazer_db.update_user_is_archived(user=existing_user, archive=False)
    LOG.info(f"User unarchived: {email}")


@base.command()
@click.argument("analysis_id", type=int)
@click.pass_context
def cancel(context, analysis_id):
    """Cancel all jobs in an analysis."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    try:
        trailblazer_db.cancel_ongoing_analysis(analysis_id=analysis_id, email=environ_email())
    except Exception as error:
        LOG.error(error)


@base.command("set-completed")
@click.argument("analysis_id", type=int)
@click.pass_context
def set_analysis_completed(context, analysis_id):
    """Set status of an analysis to 'COMPLETED'."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    try:
        trailblazer_db.update_analysis_status_to_completed(analysis_id=analysis_id)
    except Exception as error:
        LOG.error(error)


@base.command("set-status")
@click.option(
    "--status",
    is_flag=False,
    type=str,
    help=f"Status to be set. Can take the values:{TrailblazerStatus.statuses()}",
)
@click.argument("case_id", type=str)
@click.pass_context
def set_analysis_status(
    context,
    case_id: str,
    status: str,
):
    """Set the status of the latest analysis for a given case id."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    try:
        trailblazer_db.update_analysis_status(case_id=case_id, status=status)
    except ValueError as error:
        LOG.error(error)
        raise click.Abort from error
    except Exception as error:
        LOG.error(error)


@base.command()
@click.option("--force", is_flag=True, help="Force delete if analysis is ongoing")
@click.option("--cancel-jobs", is_flag=True, help="Cancel all ongoing jobs before deleting")
@click.argument("analysis_id", type=int)
@click.pass_context
def delete(context, analysis_id: int, force: bool, cancel_jobs: bool):
    """Delete analysis completely from database, and optionally cancel all ongoing analysis jobs."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    try:
        if cancel_jobs:
            trailblazer_db.cancel_ongoing_analysis(analysis_id=analysis_id)
        trailblazer_db.delete_analysis(analysis_id=analysis_id, force=force)
    except Exception as error:
        LOG.error(error)


@base.command("ls")
@click.option(
    "-s",
    "--status",
    type=click.Choice(TrailblazerStatus.statuses()),
    help="Find analysis with specified status",
)
@click.option("-b", "--before", type=str, help="Find analyses started before date")
@click.option("-c", "--comment", type=str, help="Find analysis with comment")
@click.option("--limit", type=int, default=30, help="Limit the number of analysis returned")
@click.pass_context
def ls_cmd(context, before: str, status: TrailblazerStatus, comment: str, limit: int = 30):
    """Display recent logs for the latest analyses."""
    trailblazer_db: Store = context.obj["trailblazer_db"]
    analyses: list[Analysis] | None = trailblazer_db.get_analyses_by_status_started_at_and_comment(
        status=status,
        before=datetime.strptime(before, TRAILBLAZER_TIME_STAMP).date() if before else None,
        comment=comment,
    )
    if analyses:
        for analysis in analyses[:limit]:
            (message, color) = _get_ls_analysis_message(analysis=analysis)
            click.echo(click.style(message, fg=color))
    else:
        LOG.warning("No analyses matching search criteria")



import subprocess
from pathlib import Path
from typing import Callable

from trailblazer.apps.slurm.models import SqueueResult
from trailblazer.apps.slurm.utils import formatters
from trailblazer.constants import (
    CharacterFormat,
    FileFormat,
    SlurmJobStatus,
    TrailblazerStatus,
)
from trailblazer.exc import EmptySqueueError
from trailblazer.io.controller import ReadFile, ReadStream


def _get_squeue_jobs_flag_input(slurm_job_id_file_content: dict[str, list[str]]) -> str:
    """Return a string of comma separated SLURM job ids to be used as input for squeue jobs flag."""
    job_ids: list[str] = []
    for slurm_job_ids in slurm_job_id_file_content.values():
        [job_ids.append(str(job_id)) for job_id in slurm_job_ids]
    return ",".join(job_ids)


def cancel_slurm_job(slurm_id: int, analysis_host: str | None = None) -> None:
    """Cancel SLURM job by SLURM job id."""
    scancel_commands: list[str] = ["scancel", str(slurm_id)]
    if analysis_host:
        scancel_commands = ["ssh", analysis_host] + scancel_commands
    subprocess.Popen(scancel_commands)


def _get_slurm_queue(job_ids: str, analysis_host: str | None = None) -> SqueueResult:
    """Return squeue output from ongoing analyses in SLURM."""
    queue_output: str = get_slurm_queue_output(job_ids=job_ids, analysis_host=analysis_host)
    return get_squeue_result(queue_output)


def get_slurm_squeue_output(slurm_job_id_file: Path, analysis_host: str | None = None) -> str:
    """Return squeue output from ongoing analyses in SLURM."""
    slurm_job_id_file_content: dict[str, list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=slurm_job_id_file
    )
    slurm_jobs: str = _get_squeue_jobs_flag_input(slurm_job_id_file_content)
    return get_slurm_queue_output(job_ids=slurm_jobs, analysis_host=analysis_host)


def get_slurm_queue_output(job_ids: str, analysis_host: str | None = None) -> str:
    """Return squeue output from ongoing analyses in SLURM."""
    squeue_commands: list[str] = [
        "squeue",
        "--jobs",
        job_ids,
        "--states=all",
        "--format",
        "%A,%j,%T,%l,%M,%S",
    ]
    if analysis_host:
        squeue_commands = ["ssh", analysis_host] + squeue_commands
        return (
            subprocess.check_output(
                squeue_commands,
                universal_newlines=True,
            )
            .decode(CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8)
            .strip()
            .replace("//n", "/n")
        )
    return subprocess.check_output(squeue_commands).decode(
        CharacterFormat.UNICODE_TRANSFORMATION_FORMAT_8
    )


def get_squeue_result(squeue_response: str) -> SqueueResult:
    """Return SqueueResult object from squeue response.
    Raises:
        TrailblazerError: when no entries were returned by squeue command.
    """
    if not squeue_response:
        raise EmptySqueueError("No jobs found in SLURM registry")
    squeue_response_content: list[dict] = ReadStream.get_content_from_stream(
        file_format=FileFormat.CSV,
        stream=squeue_response,
        read_to_dict=True,
    )
    return SqueueResult(jobs=squeue_response_content)


def reformat_squeue_result_job_step(data_analysis: str, job_step: str) -> str:
    """Standardise job step string according to data analysis."""
    formatter: Callable = formatters.formatter_map.get(data_analysis, formatters.reformat_undefined)
    return formatter(job_step=job_step)


def _get_analysis_single_status(jobs_status_distribution: dict[str, float]) -> str:
    """Return true if only one status in jobs statuses."""
    if len(jobs_status_distribution) == 1:
        single_status: str = jobs_status_distribution.popitem()[0]
        return (
            TrailblazerStatus.FAILED
            if single_status == SlurmJobStatus.TIME_OUT
            else TrailblazerStatus[single_status.upper()]
        )


def _is_analysis_failed(jobs_status_distribution: dict[str, float]) -> bool:
    """Return true if any job was broken."""
    return any(
        broken_status in jobs_status_distribution
        for broken_status in [SlurmJobStatus.FAILED, SlurmJobStatus.TIME_OUT]
    )


def _is_analysis_ongoing(jobs_status_distribution: dict[str, float]) -> bool:
    """Return True if analysis is still ongoing."""
    return any(
        ongoing_status in jobs_status_distribution
        for ongoing_status in TrailblazerStatus.ongoing_statuses()
    )


def get_current_analysis_status(jobs_status_distribution: dict[str, float]) -> str:
    """Return current analysis status based on jobs status distribution."""
    if single_analysis_status := _get_analysis_single_status(jobs_status_distribution):
        return single_analysis_status
    is_analysis_ongoing: bool = _is_analysis_ongoing(jobs_status_distribution)
    is_analysis_failed: bool = _is_analysis_failed(jobs_status_distribution)
    if is_analysis_failed:
        return TrailblazerStatus.ERROR if is_analysis_ongoing else TrailblazerStatus.FAILED
    if is_analysis_ongoing:
        return TrailblazerStatus.RUNNING
    return TrailblazerStatus.CANCELLED
